#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <sys/stat.h>
#include <stdarg.h>
#include <archive.h>
#include <archive_entry.h>
#include <unistd.h>
#include <fcntl.h>
#include "listfiles.h"
#include "ptrbuf.h"

#define _COMLP0  
#define _COMLP1 (*
#define _COMRP0 
#define _COMRP1 )
#define _LEFP(p) _COMLP##p
#define _RIGP(p) _COMRP##p
#define _PD(p,x) _LEFP(p) x _RIGP(p)

   // here, ft is the file entry type such as 'D', 'F' etc.
#define FILTERFUNC(x,p) \
  int _PD(p,x) (int ft,char *path,char *dir,char *file, char *arg1,char *arg2)

static char *pakdir,*fullpath;

static ptrbuf_t *preRun, *postRun;

static int hasprefix(char *str,char *pfx)
{
  int Ls,Lp;
  Ls= strlen(str);
  Lp= strlen(pfx);
  if (Ls<=Lp) return 0;
  return !strncmp(str, pfx, Lp);
}

static int hassuffix(char *str,char *sfx)
{
  int Ls, Lf;
  Ls= strlen(str);
  Lf= strlen(sfx);
  if (Ls<=Lf) return 0;
  str+= Ls - Lf;
  return !strcmp(str, sfx);
}

static void die(const char *fmt, ...)
{
  va_list ap;
  va_start(ap,fmt);
  vfprintf(stderr,fmt,ap);
  va_end(ap);
  exit(1);
}

typedef struct filter
{
  FILTERFUNC(func,1);
  char *args[2];
  struct filter *next;
} filter_t;

static FILTERFUNC(FIdot,0) { return !strcmp(path,"."); }
static FILTERFUNC(FIfalse,0) { return 0; }
static filter_t FNall= { FIdot,  NULL, NULL, NULL };
static filter_t FPall= { FIfalse, NULL, NULL, NULL };
static filter_t *FNhead=&FNall, *FNtail=&FNall;
static filter_t *FPhead=&FPall, *FPtail=&FPall;

static FILTERFUNC(FIname,0) { return !strcmp(file, arg1); }
static FILTERFUNC(FIdir,0)
{
  if (!strcmp(dir,arg1)) return 1;
  if (!hasprefix(dir,arg1)) return 0;
  return (dir[strlen(arg1)]=='/');
}
static FILTERFUNC(FIsfx,0) { return hassuffix(file, arg1); }
static FILTERFUNC(FIpfx,0) { return hasprefix(file, arg1); }
static FILTERFUNC(FIfile,0) { return !strcmp(path,arg1); }

typedef struct
{
  char *name;
  int Nmagic;
  unsigned char magic[100];
} magic_t;

static magic_t magics[]=
{
  { "elf", 4, { 0x7f, 'E', 'L', 'F' } }, 
  { "pdf", 4, { '%', 'P', 'D', 'F' } },
  { "postscript", 4, { '%', '!', 'P', 'S' } },      
  { NULL, 0, }
};

static FILTERFUNC(FImagic,0)
{
  int i,j;
  int found;
  unsigned char buf[100];
  int T,R,fd,K;

  if (ft!='F') return 0;
  found= 0;
  for(i=0;magics[i].name;i++)
  {
    R= strcmp(magics[i].name,arg1);
    if (R<0) continue;
    if (R>0) break;
    found= 1;
    T= magics[i].Nmagic;
    fd= open(path, O_RDONLY);
    if (fd<0) { fprintf(stderr,"warning: can't open %s: %s\n", path,
                                strerror(errno));
                return 0; }
    for(R=0;R<T;R+=K)
     if ((K=read(fd,buf+R,T-R))<0)
     { 
       fprintf(stderr,"warning: can't read %s: %s\n", path, strerror(errno));
       close(fd); 
       return 0; 
     }
     close(fd);
     for(j=0;j<T;j++) if (buf[j]!=magics[i].magic[j]) break;
     if (j==T) return 1;
  }
  if (!found) die("unknown magic %s\n", arg1);
  return 0;
}

static struct archive *AR;


static char *strcom(char *s1, ...)
{
  va_list ap;
  int L;
  char *R,*K;
  R= NULL;

again:
  L= strlen(s1);
  if (R) strcpy(R,s1);
  va_start(ap,s1);
  while((K= va_arg(ap, char*))) { if (R) strcpy(R+L,K); L+= strlen(K); }
  va_end(ap);
  if (R) return R;
  R= malloc(L+1);
  goto again;
}

static int lineno;

static int input_line(FILE *inp, char *line,int sz)
{
  int i,j,L;
again:
  memset(line,0, sz);
  if (!fgets(line, sz-1, inp)) return 0;
  line[sz-2]= 0;
  for(i=0;line[i];i++)
    if (line[i]=='\r' || line[i]=='\n') { line[i]=0; break; }
  lineno++;
  if (!line[0]) goto again;
  for(i=0;line[i];i++) if (line[i]<' ') line[i]= ' ';
  j= 0;
  for(i=0;line[i];i++)
  {
    if ((line[i]==' ') && (j==0 || line[j-1]==' ')) continue;
    line[j++]= line[i];
  }
  line[j]= 0;
  while(j>0 && line[j-1]==' ') line[--j]= 0;
  if (line[0]=='#' || line[0]==0) goto again;
  return 1;
}

static char *getword(char **IOall)
{
  char *k,*s;
  k= *IOall;
  if (!k[0]) return NULL;
  s= strchr(k, ' ');
  if (!s) { *IOall= k+strlen(k); return k; }
  *s= 0;
  *IOall= s+1;
  return k;
}

typedef struct
{
  int nargs;
  char *name;
  FILTERFUNC(func,1);
  int pfxpak;
   // if this flag is 1, then pakdir/ is prefixed to the argument
   // this is used for filters which are based on the directory name
} filterdef_t;

static void filter
 (int sense,filterdef_t *FD, char *line)
{
  filter_t *T;
  char *w;
  char *R;
  int i;
  T= calloc(1,sizeof(*T));
  for(i=0;i<FD->nargs;i++)
  {
    w= getword(&line);
    if (!w) 
       die("line %d: filter %s needs %d arguments\n", lineno,
                              FD->name, FD->nargs);
    if (FD->pfxpak) T->args[i]= strcom(pakdir,"/",w,NULL);
               else T->args[i]= strdup(w);
  }
  T->func= FD->func;
 
  if (sense=='-')
  {
    FNtail->next= T;
    FNtail= T;
  }
  else
  {
    FPtail->next= T;
    FPtail= T;
  }
}


static filterdef_t filterdefs[]=
{
  { 1, "name", FIname, 0 },
  { 1, "dir", FIdir, 1 },
  { 1, "sfx", FIsfx, 0 },
  { 1, "pfx", FIpfx, 0 },
  { 1, "file", FIfile, 1 },
  { 1, "magic", FImagic, 0 },
  { 0, NULL, NULL, 0 }
};

static void parse_input(char *inpname)
{
  char line[2000];
  FILE *inp;
  char *all, *word;
  int i,S;

  inp= fopen(inpname, "rb");
  if (!inp)
    die("can't open pakdef.txt: %s\n", strerror(errno));

  while(input_line(inp, line, sizeof(line)))
  {
    all= line;
    word= getword(&all);
    if (!word) continue;
    if (word[0]=='-' || word[0]=='+')
    {
      S= word[0]; 
      word++;
      for(i=0;filterdefs[i].name;i++)
        if (!strcmp(filterdefs[i].name,word))
        {
          filter(S, filterdefs+i, all);
          break;
        }
      if (!filterdefs[i].name)
         die("%s:%d: unknown filter %s\n",inpname, lineno, word);
      continue;
    }
    if (word[0]=='<' || word[0]=='>') 
    {
      S= word[0];
      word++;
      if (!strcmp(word,"run"))
      {
        if (S=='<') ptrbuf_put(preRun, strdup(all));
               else ptrbuf_put(postRun, strdup(all));
        continue;
      }
      die("%s:%d: unknown command %c%s\n",S,word);
    }
    die("%s:%d: unknown command %s\n",inpname, lineno, word);
  }
  fclose(inp);
}

static void normalize_path(char *P)
{
  int L;
  char *s,*k;
  while((L=strlen(P))) if (P[L-1]=='/') P[L-1]=0; else break;
  while((s=strstr(P,"//"))) memmove(s, s+1, strlen(s+1)+1);
  while((s=strstr(P,"/../")))
  {
    for(k=s-1;k>=P;k--) if (k[0]=='/') break;
    memmove(k+1,s+4, strlen(s+4)+1);
  }
  while((s=strstr(P,"/./"))) memmove(s,s+2, strlen(s+2)+1);
  while((s=strstr(P,"./"))==P) memmove(P, P+2, strlen(P+2)+1);
}


static int apply_filters(int ft, char *path)
{
  char *dir, *file;
  char *s;
  int R;
  filter_t *f;
  path= strdup(path);
  normalize_path(path);
  s= strrchr(path,'/');
  if (!s) { dir= strdup("."); file= path; }
  else    { dir= strdup(path); dir[s-path]= 0; file= dir+(s-path)+1; }
  R= 0;
  for(f=FNhead;f;f=f->next)
   if ((*f->func)(ft, path, dir, file, f->args[0], f->args[1]))
    { R= 1;
      for(f=FPhead;f;f=f->next)
        if ((*f->func)(ft,path,dir,file,f->args[0],f->args[1])) 
             { R= 0; break; }
      break;
    }
  free(path);
  free(dir);
  return R;
}

static void getpakdir()
{
  char wd[2000];
  int L;
  char *s;
  if (!getcwd(wd,sizeof(wd))) die("getcwd: %s\n", strerror(errno));
  L= strlen(wd);
  if (L>0 && wd[L-1]=='/') wd[L-1]= 0;
  if (!wd[0]) die("can't run at the root directory.\n");
  s= strrchr(wd,'/');
  if (!s) die("getpakdir: bad directory(%s)\n", wd);
  pakdir= strdup(s+1);
  printf("package name is %s.tar.bz2\n",pakdir);
}

static const char *aerr()
{
  return archive_error_string(AR);
}

static void begin_archive(char *tdir)
{
  char *K;
  AR= archive_write_new();
  if (!AR) die("libarchive internal error in _new()\n");
  if (archive_write_set_format_ustar(AR) ||
      archive_write_add_filter_bzip2(AR))
    die("libarchive: can't set format .tar.bz2: %s\n", aerr());
  K= strcom(tdir,"/",pakdir,".tar.bz2", NULL);
  if (archive_write_open_filename(AR,K))
    die("libarchive_open: %s\n", aerr());
  fullpath= K;
}

static void add_dir(file_ent_t *fe)
{
  struct archive_entry *ae;

  ae= archive_entry_new();
  if (!ae) die("libarchive: error in entry_new(): %s\n",aerr());

  printf("DIR [%s]\n", fe->path);

  archive_entry_copy_stat(ae, &fe->sbuf);
  archive_entry_copy_pathname(ae, fe->path);
  if (archive_write_header(AR,ae))
    die("libarchive: write_header(): %s\n",aerr());
  archive_entry_free(ae);
}

static void add_link(file_ent_t *fe)
{
  struct archive_entry *ae;
  char buf[2000];
  int sz,r;

  sz= sizeof(buf);
  r= readlink(fe->path, buf, sz-1);
  if (r<0) die("readlink(%s): %s\n", fe->path, strerror(errno));
  buf[r]= 0;

  ae= archive_entry_new();
  if (!ae) die("libarchive: error in entry_new(): %s\n",aerr());

  printf("LNK [%s]\n", fe->path);

  archive_entry_copy_stat(ae, &fe->sbuf);
  archive_entry_copy_pathname(ae, fe->path);
  archive_entry_copy_symlink(ae, buf);
  
  if (archive_write_header(AR,ae))
    die("libarchive: write_header(): %s\n",aerr());
  archive_entry_free(ae);
}

static void add_file(file_ent_t *fe)
{
  struct archive_entry *ae;
  int fd;
  char buf[4*512],*p;
  int L,R;

  ae= archive_entry_new();
  if (!ae) die("libarchive: error in entry_new(): %s\n",aerr());

  printf("FIL [%s]\n", fe->path);

  archive_entry_copy_stat(ae, &fe->sbuf);
  archive_entry_copy_pathname(ae, fe->path);
  fd= open(fe->path, O_RDONLY);
  if (fd<0) die("open(%s): %s\n", fe->path, strerror(errno));
  if (archive_write_header(AR,ae))
    die("libarchive: write_header(): %s\n",aerr());
  while((L=read(fd, buf, sizeof(buf)))>0)
  {
    for(p=buf;L;p+=R,L-=R)
     if ((R=archive_write_data(AR,p,L))<0)
       die("libarchive: write_data(): %s\n",aerr());
  }
  close(fd);
  archive_entry_free(ae);
}

static void run_commands(ptrbuf_t *cc)
{
  char *C;
  while((C=ptrbuf_get(cc)))
  {
    printf("RUN [%s]\n",C);
    if (system(C)) break;
  }
}

int main(int argc,char **argv)
{
  char *tdir;
  file_ent_t *fe;
  preRun= ptrbuf_new(50);
  postRun= ptrbuf_new(50);
  if (argc!=2) die("usage: %s directory-for-tarbz2\n", argv[0]);
  getpakdir();
  tdir= argv[1];
  parse_input("pakdef.txt");
  run_commands(preRun);
  if (chdir("..")) die("can't chdir(..): %s\n", strerror(errno));
  begin_archive(tdir);
  for(fe=listfilesBegin(pakdir);fe;fe=listfilesNext(fe))
  {
    if (apply_filters(fe->type, fe->path)) continue;
    switch(fe->type)
    {
    case 'D': add_dir(fe); break;
    case 'F': add_file(fe); break;
    case 'L': add_link(fe); break;
#if 0
    case 'U': unknown
    case 'C': character device
    case 'B': block device
    case 'P': fifo
    case 'S': socket
#endif
    default:
      continue;       
    }
  }
  if (archive_write_close(AR))
    die("libarchive: write_close(): %s\n", aerr());
  archive_write_free(AR);
  printf("archive written to %s\n",fullpath);
  if (chdir(pakdir)) die("can't chdir(%s): %s\n", pakdir,strerror(errno));
  run_commands(postRun);
  return 0;
}

